/* $Id: counter.c,v 2.1 1995/06/24 19:37:15 ssd Exp ssd $ */
/*
 *  counter.c
 *      Eric C. Rosen and Steven Dick
 *      Copyright May, 1995
 * Copy freely; send e-mail to insure that you get notices of updates.
 * Please let me know of any interesting changes you make.
 *
 * Steven Dick: ssd@mae.engr.ucf.edu
 *      http://www-mae.engr.ucf.edu/~ssd/counter.html
 * Eric Rosen:  eric@cse.ucsc.edu
 *      http://sapphire.cse.ucsc.edu/cgi-bin/counter
 */

/* bugs:
 *   portability concerns:
 *     ndbm not available on all platforms (try dbm? -ldbm? etc.)
 *     lockf not available (fcntl, flock, semaphore)
 *     not all systems use unix style time()
 *     tags longer than 255 will bomb
 *     the code to pack and unpack the URL database entries is very ugly
 * todo:
 *   write a quick perl script template to read the database
 *   add support for nph- use
 *   add alarm call to kill stuck sockets after some reasonable timeout
 *   color
 *
 * UNTESTED:
 *   lockf
 */

#include <stdio.h>
#include <stdlib.h>
#include <ndbm.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>		/* gmtime */
#include <sys/stat.h>

#include <sys/file.h>		/* flock, lockf, fcntl */

#include <limits.h>
#if 0
#include <db.h>
#endif

#ifdef macII
FILE  *popen(const char *__command, const char *__type);
#endif

#include "counter.h"
#include "jerk.h"
#include "gifencode.h"

/***********************************************************************
 *  SITE SPECIFIC CONFIG OPTIONS:
 ***********************************************************************/
/* path options */
#if 1
#define DBbase	    "/usr/datlib/private/www/logs/c-"
#else
/* for debugging */
#define DBbase	    "/tmp/"
#endif
#define DB_COUNT    DBbase "counters"
#define DB_URL      DBbase "counter_urls"
#define HOST        "www-mae.engr.ucf.edu"
char *Doc_url =     "/~ssd/counter.html";

#define DB_PERMS 0644
#define MAX_URL  60    /* truncate URL to MAX_URL; undefine to keep whole */

#ifdef NOSTRCASECMP
/* if strcasecmp not available, just do case sensitive compares */
#define strcasecmp strcmp
#define strncasecmp strncmp
#endif

/*
 * uncomment exactly one of these
 */
/* #define USE_FCNTL */		/* POSIX */
#define USE_LOCKF		/* Solaris 2 */
/*#define USE_FLOCK*/		/* */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* implementation of various locking methods */
#ifdef USE_LOCKF
#define LOCK_DECL
#define DO_LOCK(fd)   lockf(fd, F_LOCK,1);
#define DO_UNLOCK(fd) lockf(fd, F_ULOCK,1);
#endif

#ifdef USE_FLOCK
#define LOCK_DECL
#define DO_LOCK(fd)   flock(fd, LOCK_EX)
#define DO_UNLOCK(fd) flock(fd, LOCK_UN)
#endif

#ifdef USE_FCNTL
#define LOCK_DECL     struct flock fls
#define DO_LOCK(fd)   fls.l_type = F_WRLCK; fls.l_whence = SEEK_SET; \
                      fls.l_start=0; fls.l_len=1; fcntl(fd, F_SETLKW,&fls);
#define DO_UNLOCK(fd) fls.l_type = F_UNLCK; fls.l_whence = SEEK_SET; \
                      fls.l_start=0; fls.l_len=1; fcntl(fd, F_SETLKW,&fls);
#endif

/* end of config section */
/***********************************************************************/


/*
 * opcodes:
 */

#define COUNTER_INVALID        -1
#define COUNTER_VALUE           1
#define COUNTER_SHOW            1 /* mask */
#define COUNTER_INCREMENT       2 /* mask */
#define COUNTER_INCREMENTSHOW   3
#define COUNTER_NUMBER          5
#define COUNTER_DIGITS          9
#define COUNTER_RANDOM          17

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

struct font_info *Font;
char *DigitChars;
int getpixel(int x, int y)
{
  return Font->digits[DigitChars[x/Font->x]-'0'][x%Font->x+y*Font->x];
}

long Now;  /* timestamp */

/*
 * a 1-bit 1x1 GIF
 */
unsigned char   tiny[] = {
        0x47, 0x49, 0x46, 0x38, 0x37, 0x61, 0x01, 0x00,
        0x01, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x2C, 0x00, 0x00, 0x00, 0x00,
        0x01, 0x00, 0x01, 0x00, 0x00, 0x02, 0x00, 0x44,
        0x01, 0x00, 0x3B, 0x00
};

/*
 *  print_tinyGIF():
 *
 *      spit out the empty GIF above
 */
void    print_tinyGIF()
{
    printf("Content-type: image/gif\n\n");
    fwrite(tiny,1,sizeof(tiny),stdout);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Expire this url some number of seconds into the future
 */
char *Days[] = { "Sun", "Mon" , "Tue", "Wed", "Thu", "Fri", "Sat" };
char *Months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun","Jul","Aug","Sep",
		   "Oct","Nov","Dec"};
void expire(long secs)
{
  long when;
  struct tm *t;

  /* calculate the time string */
  when = time(0) + secs;  /* expire in an hour */
  t = gmtime(&when);
  printf("Expires: %s, %02d %s %d %02d:%02d:%02d GMT\n", Days[t->tm_wday],t->tm_mday,Months[t->tm_mon], t->tm_year+1900, t->tm_hour, t->tm_min, t->tm_sec);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 *  print_number():
 *
 *      spit out a GIF image of for the indicated number
 *
 *      if width is positive and non-zero, pad field with zeros to the
 *      width
 */
void    print_number(long n, int width, struct font_info *f, long exp)
{
  char    nbuf[128],  fbuf[48];
  int w;

  if (n<0) return;
  fprintf(stdout, "Content-type: image/gif\n");
  if (exp) expire(exp);
  putchar('\n');
  fflush(stdout);

  if (!width || width>120)
    strcpy(fbuf, "%0ld");
  else
    sprintf(fbuf, "%%0%dld", width);
  sprintf(nbuf, fbuf, n);
  w = strlen(nbuf);

#if 0
  {
    FILE    *pF = NULL;
    int     j, k;
    /* if using ppmtogif instead of internal gifencode */
    pF = popen(PPMTOGIF, "w");

    /* print OUT "P2\n",15*(1+$#digits), " 20\n255\n"; */
    fprintf(pF, "P5\n%d %d\n255\n", f->x*w, f->y);
    for (j=0; j<f->y; j++) {
      for (k=0; k<w; k++) {
	fwrite(f->digits[nbuf[k] - '0']+f->x*j,1,f->x,pF);
      }
    }
    pclose(pF);
  }
#else
  Font = f;
  DigitChars = nbuf;
  GIFEncode(stdout, f->x*w,f->y,0,0,-1,f->nbits,f->greymap,f->greymap,f->greymap,getpixel);
#endif
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Return an error
 *
 * Also, tell the client that the error won't go away. [i.e. cache it]
 */
void badkey(char *why,char *part)
{

#if 0
  /* for npl- stuff */
  printf("HTTP/1.0 404 Counter Error: %s  -- please read http://%s%s\n",why, HOST,Doc_url);
#endif
  printf("Content-type: text/html\n");
  expire(3600);
  putchar('\n');
  printf("<TITLE>counter error!</TITLE>\n");
  printf("Bad handle -- please read the <a href=\"http://%s%s\">counter documnetation</a>\n<p>Error: <b>%s",HOST,Doc_url,why);
  if (part) printf(" '<tt>%s</tt>'",part);
  printf("</tt></b>\n");
  exit(0);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * remove all http escapes from a string, destroying the original;
 * stop at char c or '\0'; return 0 if finished or zero char matching c
 * and return location after c.
 * Also check for illegal characters.
 */
char* unescape(char *s,char c)
{
  char *t;
  for (t=s; *s && *s!=c; s++,t++) {
    if (*s=='%'){
      s++;
      c = *s-'0';
      if (*s>'9') c-= 'A'-'0'+10;
      if (*s>='a') c-= 'a'-'A';
      s++;
      c *= 16;
      c += *s-'0';
      if (*s>'9') c-= 'A'-'0'+10;
      if (*s>='a') c-= 'a'-'A';
      if (c <= 32) badkey("Illegal char in URL",0);
      *t=c;
    } else if (t!=s)
      *t = *s;
    if (*t<=32)
      badkey("Illegal char in URL",0);
  }
  if (*s) {
    *s=0; 
    return ++s;
  }
  else return 0;
}

/*
 * determine what should be done...
 * argument taken from $ENV{QUERY_STRING} rather than command line if possible
 */
int parse_arg(char *cmdline, char **tagp,int *widthp, struct font_info **fontp)
{
    static char *badtags[] ={
      "index.html", "index", "welcome", "welcome.html", "number", "value",
      "increment", "show", "noshow", 0};
    char *c,*s,*t, **sp;
    int opcode=COUNTER_INCREMENTSHOW; /* default opcode */

    /* should POST method be suported? */
    c = getenv("QUERY_STRING");	/* GET method */
    if (!c) c=cmdline;	 /* retain this to make cmd line  debugging easier */
    s = strchr(c,'&');

    if (s) {
      *s = '\0';
    }
    *tagp = c;
    if (unescape(c,'='))
      badkey("Illegal char '=' found in key '%s'.",c);

    /* only these values are allowed for empty tags */
    if (!strcasecmp(c, "digits")) {
      opcode= COUNTER_DIGITS;
      *tagp=0;
    } else if (!strcasecmp(c, "random") || !strcasecmp(c,"rand")) {
      opcode= COUNTER_RANDOM;
      *tagp=0;
    } else if (*tagp && isdigit(**tagp))
      opcode = COUNTER_NUMBER;

    if (*tagp) {
      /* check for worthless tags */
      /* font= and width= should be caught above */
      for (sp = badtags; *sp; sp++)
	if (!strcasecmp(*tagp,*sp))
	  badkey("Ambiguous tag",*tagp);
    }

    while (s) {			/* look for other options */
      char *v;
      s++;		     /* skip the '&' */
      t = strchr(s,'&');     /* search for the next '&' */
      if (t) *t=0;
      v=unescape(s,'=');

      if (v && !strcasecmp(s,"width")) {
	*widthp = atoi(v);
	opcode |=COUNTER_SHOW;
      } else if (v && !strcasecmp(s,"font")){
	struct font_info *f;
	for (f=FontList; f->fontname; f++)
	  if (!strcasecmp(v,f->fontname))
	    break;
	if (f->fontname)
	  *fontp = f;
	else
	  badkey("Unknown font",v);
	opcode |= COUNTER_SHOW;
      } else if (!v && *tagp) { /* these options require a tag and no value */
	if (!strcasecmp(s, "number") || isdigit(**tagp))
	  opcode = COUNTER_NUMBER;
	else if (!strcasecmp(s, "value") || !strcasecmp(s, "noinc"))
	  opcode = COUNTER_VALUE;
	else if (!strcasecmp(s , "increment"))
	  opcode = COUNTER_INCREMENTSHOW;
	else if (!strcasecmp(s,"show"))
	  opcode |= COUNTER_SHOW;
	else if (!strcasecmp(s, "noshow"))
	  opcode = COUNTER_INCREMENT;
	else
	  badkey("Unknown option",s);
      } else
	badkey("Unknown option",s);

      s = t; /* check the next option */
    }
    return opcode;      /* unrecognized */
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 *  tag database functions
 *
 *  Each value is a struct counter_stat.
 */

/*
 *  gettag():
 *
 *      return value of indicated tag
 *      if tag does not exist, return zero as its value
 *      increment value if desired
 *      handle locking and refererer url data
 */
 /* URL database format:
  *   long urlcount;
  *   char ntags;
  *   char lengths[ntags];
  *   char strings[ntags][lengths[i]];   (sortof)
  */
long    gettag(char *tag,int inc)
{
  DBM *db;
  int lockfile;
  datum key, value;
  struct counter_stat cs;
  char *refurl;
  LOCK_DECL;

  if (!tag) return 0;
/*  printf("tag='%s'\n",tag);*/
  key.dptr = tag;
  key.dsize = strlen(tag);

  lockfile= open(DB_COUNT ".lock", O_RDWR | O_CREAT, 0644);
  if (lockfile<0) badkey("Lockfile open failed!",0);
  DO_LOCK(lockfile);
  db = dbm_open(DB_COUNT, O_RDWR | O_CREAT , DB_PERMS);
  if (!db) badkey("Database open failed!",0);
  value = dbm_fetch(db,key);
  if (!value.dptr || value.dsize<sizeof(struct counter_stat)){
    if (value.dptr && value.dsize>=sizeof(long))
      memcpy(&cs.count,&value.dptr,sizeof(long)); /* keep old count, reinit rest */
    else
      cs.count=0;
    cs.weekly=1;    cs.lastreset=0;
    cs.creation = cs.lastaccess = Now;
  } else { /* should also check for size!! */
    memcpy(&cs,value.dptr,sizeof(struct counter_stat));
    cs.weekly++;
    cs.lastaccess = Now;
  }
  value.dptr = (void*)&cs;
  value.dsize = sizeof(struct counter_stat);
  if (inc && cs.count>=0) cs.count++;
  dbm_store(db,key,value,DBM_REPLACE );
  dbm_close(db);
  DO_UNLOCK(lockfile);
  close(lockfile);
  refurl = getenv("REFERER_URL");
  if (refurl && *refurl){	                      /* update URL info */
    char *s;
    lockfile= open(DB_URL ".lock", O_RDWR | O_CREAT, 0644);
    if (lockfile<0) return cs.count;     /* just pass */
    DO_LOCK(lockfile);
    db = dbm_open(DB_URL, O_RDWR | O_CREAT , DB_PERMS);
    if (!db) return cs.count;		 /* just pass */
    key.dptr = refurl;
    key.dsize = strlen(refurl);
    if ((s=strrchr(refurl,'#')) && (s-refurl>key.dsize))
      key.dsize = s-refurl;		 /* chop off internal anchor */
#ifdef MAX_URL
    if (key.dsize>60) key.dsize = MAX_URL;
#endif
    value = dbm_fetch(db,key);
    if (!value.dptr || value.dsize<2+sizeof(long)){	 /* make a new entry */
      value.dptr = s = (char*)malloc( value.dsize = strlen(tag)+2+sizeof(long));
      *((long*)s) = 1;
      s += sizeof(long);
      s[0]=1;		   /* one entry */
      s[1]=strlen(tag);
      memcpy(s+2,tag,s[1]);
      dbm_store(db,key,value,DBM_REPLACE );
    } else {				 /* update an old entry */
      int i,ss;
      unsigned char *t;
      ss = strlen(tag);
      /* search to see if key is already there */
      for (i=0, t=value.dptr+sizeof(long),s=t+t[0]+1; i<t[0]; s+=t[1+i++]){
        /*fprintf(stderr,"test[%d]: %2x%2x %d %2x%2x %d '%.*s'\n",i,*tag,tag[ss-1],ss,*s,s[t[i+1]-1],t[i+1],t[i+1],s); DEBUG*/
	if (!strncmp(tag,s,t[i+1]))
	  break;
      }
      if (i>=t[0]){
	/* tag wasn't found -- add it to end */
	int newsize;
	long *lp;
	/*fprintf(stderr,"adding %d %s\n",ss,tag); DEBUG */
	s = (char*)malloc( newsize = value.dsize +ss+1);
	lp = (long*)s;
	memcpy(s,value.dptr,t[0]+sizeof(long)+1); /* copy lengths and count */
	value.dptr = s;
	++*lp;
	s+=sizeof(long);  /* done with first field */
	value.dsize -= sizeof(long);
	++*s;  /* one more tag... */
	s[*s]=ss;  /* save the new length */
	memcpy(s+*s+1,t+*s,value.dsize-*s); /* copy old tags */
	memcpy(s+value.dsize+1,tag,ss);             /* copy new tag */
	value.dsize = newsize;
	dbm_store(db,key,value,DBM_REPLACE );
      } else { /* just increment the value */
	long *lp;
	s = (char*)malloc(value.dsize);
	lp = (long*)s;
	memcpy(s,value.dptr,value.dsize);
	++*lp;
	value.dptr = s;
	dbm_store(db,key,value,DBM_REPLACE );
      }
    }
    dbm_close(db);
    DO_UNLOCK(lockfile);
    close(lockfile);
  }
  return cs.count;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int main(int argc, char *argv[])
{
    struct font_info *f=&FontList[0];
    int width=0;

    char    *tag = NULL;
    long    value = 0L;
    int     i, opcode;

    /* set the timestamp so all stamps can be identical */
    Now = time(0);

    /*
     * parse argument and process:
     */
    opcode = parse_arg(argv[1], &tag,&width,&f);
    value = 0;
    switch (opcode) {
        case COUNTER_INCREMENT:
            value = gettag(tag,1);
            if (value>= 0)
              print_tinyGIF();
        break;

        case COUNTER_INCREMENTSHOW:
            value = gettag(tag,1);
            print_number( value, width, f, 0 );
        break;

        case COUNTER_VALUE:
            value = gettag(tag,0);
            print_number( value, width, f, 0 );
        break;

        case COUNTER_DIGITS:
            print_number(1234567890L, 10, f, 7000);
        break;

        case COUNTER_RANDOM:
            srand(time(0));
            i = rand();
            print_number( (long) i, width, f, 20);
        break;

        case COUNTER_NUMBER:
            value=atoi(tag); /* sscanf(tag, "%ld", &value);*/
            print_number( value, width, f, 7000 );
        break;


        default:
	    { char buf[10];
	      sprintf(buf,"%d",opcode);
	      badkey("Unrecognized opcode",buf);
            }
        break;
    }
    if (value<0) {
        if (value < -1 ){
           fprintf(stdout, "Content-type: image/gif\n");
           expire(7000);
           putchar('\n');
           fwrite(jerk,1,jerk_size,stdout);
        } else { /* emit broken gif with an expire */
           fprintf(stdout, "Content-type: image/gif\n");
           expire(7000);
           putchar('\n');
        }
    }
    return(0);
}
