/* $Id: digits2c.c,v 2.1 1995/06/24 19:37:15 ssd Exp ssd $ */
/*  Copyright May, 1995 by Steven Dick <ssd@mae.engr.ucf.edu>
 *  Copy freely; send me e-mail so I can inform you of updates;
 *  let me know of any interesting changes you make.
 */

/* this handles grey scale images only */

#include <stdio.h>
#include <string.h>
#include <malloc.h>

int F_background = -1;
int F_transparent = -1;
unsigned char F_bitmask = 255;

struct image_format {
  unsigned char *data;
  int xw,yw,npix;
} Digits[10];

unsigned char colormap[256];
unsigned char colormap_real[256];

void printdata(FILE *f,unsigned char *data, int len)
{
  int c=81;

  fprintf(f,"%d",*data);
  c-=3;
  if (*data>9) c--;
  if (*data>99) c--;
  data++;
  len--;
  while (len){
    while (len && c>5){
      if (c==81)
	fprintf(f,"%d",*data);	/* first in row */
      else
	fprintf(f,", %d",*data);
      c-=3;
      if (*data>9) c--;
      if (*data>99) c--;
      data++;
      len--;
    }
    c=81;
    if (len)
      fprintf(f,",\n");
    else
      putc('\n',f);
  }
}

void
usage()
{
  fprintf(stderr,"digits2c [options] basefilename%%d identifier\n");
  fprintf(stderr," options:\n\t-background rrggbb\n\t-n nbits\n");
  fprintf(stderr," pipe characters allowed in basefilename\n"
	  " basefilename must contain a %%d\n");
  exit(1);
}

int
main(int argc, char *argv[])
{
  char *basefilename,*identifier;
  char dopipe=0;
  FILE *in, *fc, *fh;
  char buf[1000];
  int i,j,k,xw,yw,nc;
  unsigned char *b;
  char h1,h2;

  { char **ap;
#define checkarg if (!*++ap) usage()
    for (ap=argv+1; *ap && **ap=='-'; ap++){
      switch(ap[0][1]){
       case 'b':		/* background color */
	checkarg;
	F_background=atoi(*ap);
	break;
       case 't':		/* transparent color */
	checkarg;
	F_transparent=atoi(*ap);
	break;
       case 'n':		/* number of bits to keep in colors */
	checkarg;
	F_bitmask &= ~((1<<(8-atoi(*ap)))-1);
	printf("bitmask=%x\n",F_bitmask);
	break;
       default:
	fprintf(stderr,"Unknown option %s\n",*ap);
	usage();
      }
#undef checkarg
    }
    if (!*ap) usage();
    basefilename = *ap;
    if (!*++ap) usage();
    identifier = *ap;
    if (strchr(basefilename,'|')) dopipe=1;
    if (*++ap) fprintf(stderr,"extraneous arguments ignored. (%s...)\n",*ap);
  }

  /* mark all colors as unsed */
  for (i=0;i<256;i++)
    colormap[i]=0;

  for (i=0;i<10;i++){
    sprintf(buf,basefilename,i);
    if (dopipe)
      in = popen(buf,"r");
    else
      in = fopen(buf,"r");
    if (!in) {
      perror(buf);
      exit(1);
    }

    /* read pgm header */
    j=fscanf(in,"%c%c\n%d %d %d\n",&h1,&h2,&xw,&yw,&nc);
    if (h1!='P' || h2!='5'){
      fprintf(stderr,"Not pgm binary?!?\n");
      exit(1);
    }
    if (j!=5){
      fprintf(stderr,"Error reading pgm header--only %d/5 fields read.\n",j);
      exit(1);
    }
    Digits[i].xw = xw;
    Digits[i].yw = yw;
    Digits[i].data = b = malloc(j=xw*yw);
    Digits[i].npix = j;

    if (Digits[0].xw != xw || Digits[0].yw !=yw){
      fprintf(stderr,"Oops!! Digit %d is %dx%d instead of %dx%d!\n",i,xw,yw,Digits[0].xw,Digits[0].yw);
      exit(1);
    }

    if (!b){
      fprintf(stderr,"Out of memory!\n");
      exit(1);
    }
    while (!feof(in) && j){
      k = fread(b,1,j,in);
      if (k<0){
	perror("read");
	exit(1);
      }
      j -= k;
      b += k;
      if (j) printf(" %d \r",j);
    }
    for (b=Digits[i].data, j=Digits[i].npix;j ; b++,j--)
      colormap[*b] = 1;

    if (dopipe)
      pclose(in);
    else
      fclose(in);
  }

  /* make the colormap */
  for (i=j=0; i<256;i++)
    if (colormap[i]){
      colormap_real[j]=i&F_bitmask;
      if (!j || colormap_real[j-1]!=colormap_real[j]) {
	colormap[i] = j;
	j++;
      } else
	colormap[i] = j-1;
    }
/*  printdata(stdout,colormap,256);*/

  sprintf(buf,"%s.c",identifier);
  fc = fopen(buf,"w");
  if (!fc){
    perror(buf);
    exit(1);
  }
  sprintf(buf,"%s.h",identifier);
  fh = fopen(buf,"w");
  if (!fh){
    perror(buf);
    exit(1);
  }

  { unsigned mask;
    for (mask=j-1, i=0; mask; mask>>=1, i++)
      ;
  }

  fprintf(fh,
	  "#define %s_width\t%d\n"
	  "#define %s_height\t%d\n"
	  "#define %s_maxcolor\t%d\n"
	  "#define %s_nbits\t%d\n\n"
	  "extern int %s_greymap[];\n\n"
	  "#define %s_red %s_greymap\n"
	  "#define %s_green %s_greymap\n"
	  "#define %s_blue %s_greymap\n\n",

	  identifier,Digits[0].xw,
	  identifier,Digits[0].yw,
	  identifier,j,
	  identifier,i,
	  identifier,  identifier,identifier,  identifier,identifier,  
	  identifier,identifier );

/* adjust the colormap for 8 bits */
  if (nc<255) {
    nc = 255/nc;
    for (i=0;i<j; i++)
      colormap_real[i] *= nc;
  }

  fprintf(fc,"int %s_greymap[256] = {\n",identifier);
  printdata(fc,colormap_real,j);
  fprintf(fc,"};\n\n");

  for (i=0;i<10;i++){
    for (b=Digits[i].data, j=Digits[i].npix;j ; b++,j--)
      *b = colormap[*b];
    fprintf(fh,"extern unsigned char %s%d[];\n",identifier,i);
    fprintf(fc,"unsigned char %s%d[] = {\n",identifier,i);
    printdata(fc,Digits[i].data,Digits[i].npix);
    fprintf(fc,"};\n\n");
  }
  return 0;
}
